/* edit.c
 *
 * Copyright (C) 1999 Iain Holmes <ih@csd.abdn.ac.uk>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <config.h>
#include <gnome.h>
#include <gtkpastebuffer.h>

#include "edit.h"
#include "app.h"

#define OUTLINE_WIDTH 3.0
#define SQUARE_SIZE 17.0

extern GtkObject *paste_buffer;

void
ie_selection_destroy (GtkWidget *app)
{
  IESelection *selection;
  AppMenus *am;

  selection = gtk_object_get_data (GTK_OBJECT (app), "selection");
  if (selection == NULL)
    return;

  /* Remove the data, before we kill it */
  gtk_object_set_data (GTK_OBJECT (app), "selection", NULL);

  if (selection->selitem != NULL)
    gtk_object_destroy (GTK_OBJECT (selection->selitem));

  g_free (selection);
  am = gtk_object_get_data (GTK_OBJECT (app), "menu-widgets");
  gtk_widget_set_sensitive (am->cut, FALSE);
  gtk_widget_set_sensitive (am->copy, FALSE);
  gtk_widget_set_sensitive (am->b_cut, FALSE);
  gtk_widget_set_sensitive (am->b_copy, FALSE);
  gtk_widget_set_sensitive (am->paste_into, FALSE);
}

IESelection*
ie_selection_get (GtkWidget *app)
{
  IESelection *selection;
  AppMenus *am;

  selection = gtk_object_get_data (GTK_OBJECT (app), "selection");
  if (selection == NULL)
    {
      selection = g_new0 (IESelection, 1);
      gtk_object_set_data (GTK_OBJECT (app), "selection", selection);
      /* Set the copy/cut options sensitive */
      am = gtk_object_get_data (GTK_OBJECT (app), "menu-widgets");
      gtk_widget_set_sensitive (am->cut, TRUE);
      gtk_widget_set_sensitive (am->copy, TRUE);
      gtk_widget_set_sensitive (am->b_cut, TRUE);
      gtk_widget_set_sensitive (am->b_copy, TRUE);
      if (gtk_paste_buffer_get_size (GTK_PASTE_BUFFER (paste_buffer)) > 0)
	gtk_widget_set_sensitive (am->paste_into, TRUE);
    }

  return selection;
}

static void
ie_selection_draw (GtkWidget *app)
{
  gdouble x1, y1;
  gdouble x2, y2;
  Grid *grid;
  IESelection *selection;
  GnomeCanvasGroup *group;

  g_return_if_fail (selection != NULL);

  selection = gtk_object_get_data (GTK_OBJECT (app), "selection");
  grid = gtk_object_get_data (GTK_OBJECT (app), "grid");

  if (selection->selitem != NULL)
    gtk_object_destroy (GTK_OBJECT (selection->selitem));

  x1 = selection->x * 10;
  y1 = selection->y * 10;
  x2 = (selection->x + selection->width) * 10;
    y2 = (selection->y + selection->height) * 10;

  group = GNOME_CANVAS_GROUP (GNOME_CANVAS (grid->canvas)->root);
  selection->selitem = gnome_canvas_item_new (group,
					      gnome_canvas_rect_get_type (),
					      "x1", x1,
					      "y1", y1,
					      "x2", x2,
					      "y2", y2,
					      "fill_color", NULL,
					      "outline_color", "white",
					      "fill_color_rgba", 0x8888FF55,
					      "width_pixels", 1,
					      NULL);
  gtk_widget_queue_resize (grid->canvas); /* Work round a bug in the canvas */
}

void
edit_clear_cb (GtkWidget *widget,
	       gpointer data)
{
  GtkWidget *app = GTK_WIDGET (data);
  Grid *g;
  guchar *pixels;
 
  g = gtk_object_get_data (GTK_OBJECT (app), "grid");

  pixels = gdk_pixbuf_get_pixels (g->pixbuf);
  memset(pixels, 0x00, g->rowstride*g->height); /* black */

  ie_app_dirty (app);
  ie_update_grid (app);
}

void
edit_select_all_cb (GtkWidget *widget,
		    gpointer data)
{
  GtkWidget *app = GTK_WIDGET (data);
  IESelection *selection;
  Grid *g;

  g = gtk_object_get_data (GTK_OBJECT (app), "grid");
  selection = gtk_object_get_data (GTK_OBJECT (app), "selection");

  if (selection == NULL)
    {
      selection = ie_selection_get (app);
    }

  selection->grid = g;
  selection->x = 0;
  selection->y = 0;
  selection->width = g->width;
  selection->height = g->height;

  ie_selection_draw (app);
}

static void
edit_cutcopy (GtkWidget *widget,
	      gpointer app,
	      gboolean cut)
{
  int y;
  IEPasteBuffer *pbuffer;
  IESelection *selection;
  Grid *g;
  guchar *pixels, *pixel;
  guchar *buf, *pdata;
  GdkPixbuf *pb;

  selection = gtk_object_get_data (GTK_OBJECT (app), "selection");
  if (selection == NULL)
    return;

  pbuffer = g_new (IEPasteBuffer, 1);

  g = selection->grid;

  if (cut)
    ie_app_dirty (app);

  pixels = gdk_pixbuf_get_pixels (g->pixbuf);

  pbuffer->x = selection->x;
  pbuffer->y = selection->y;
  pbuffer->width = selection->width;
  pbuffer->height = selection->height;

  pbuffer->rowstride = (selection->width * 4);
  g_assert (pbuffer->rowstride % 4 == 0);

  buf = pdata = g_new (guchar, pbuffer->rowstride * pbuffer->height);

  pixel = pixels + (pbuffer->y * g->width * 4) + (pbuffer->x * 4);

  for (y = 0; y < selection->height; y++)
    {
      /* Copy the data to the Paste Buffer, then clear it */
      memcpy (pdata, pixel, selection->width * 4);
      if (cut)
	memset (pixel, 0x00, selection->width * 4);
      
      pixel += g->rowstride;
      pdata += pbuffer->rowstride;
    }

  /* Make a gdk-pixbuf */
  pb = gdk_pixbuf_new_from_data (buf, ART_PIX_RGB, TRUE,
				 pbuffer->width, pbuffer->height,
				 pbuffer->rowstride, NULL, NULL);
  pbuffer->pixbuf = pb;

  /* Add our new IEPasteBuffer structure to the global paste buffer */
  gtk_paste_buffer_add (GTK_PASTE_BUFFER (paste_buffer), pbuffer);

  /* Remove the selection, and grey all related buttons */
  if (cut)
    ie_selection_destroy (app);

  ie_update_grid (app);
}

void
edit_cut_cb (GtkWidget *widget,
	     gpointer app)
{
  edit_cutcopy (widget, app, TRUE);
}

void
edit_copy_cb (GtkWidget *widget,
	     gpointer app)
{
  edit_cutcopy (widget, app, FALSE);
}

void
edit_paste_cb (GtkWidget *widget,
	       gpointer data)
{
  GtkWidget *app = GTK_WIDGET (data);
  IEPasteBuffer *pbuffer;
  Grid *g;
  guchar *pixels, *pixel;
  guchar *ppixels;
  gpointer kt;
  gboolean keep_trans = FALSE;
  gint y, x;

  if (gtk_paste_buffer_get_size (GTK_PASTE_BUFFER (paste_buffer)) == 0)
    {
      gnome_error_dialog_parented (_("Paste Buffer is empty"), GTK_WINDOW (app));
      return;
    }

  pbuffer = gtk_paste_buffer_get (GTK_PASTE_BUFFER (paste_buffer), 
				  GTK_PASTE_BUFFER (paste_buffer)->selection);

  kt = gtk_object_get_data (paste_buffer, "keep-trans");
  if (kt)
    keep_trans = GPOINTER_TO_INT (kt);

  g = gtk_object_get_data (GTK_OBJECT (app), "grid");

  pixels = gdk_pixbuf_get_pixels (g->pixbuf);
  pixel = pixels + (pbuffer->y * g->width * 4) + (pbuffer->x * 4);

  ppixels = gdk_pixbuf_get_pixels (pbuffer->pixbuf);

  /* There are two different ways to do this which depend on keep_trans
     because the !keep_trans way is much faster than the keep_trans way
     and I don't want to impose the speed penalty if !keep_trans */
  if (!keep_trans)
    {
      /* Fast */
      for (y = 0; y < MIN(pbuffer->height, g->height); y++)
	{
	  memcpy (pixel, ppixels, MIN(pbuffer->width, g->width) * 4);
	  
	  ppixels += pbuffer->rowstride;
	  pixel += g->rowstride;
	}  
    }
  else
    {
      /* S...l....oooooooooo.........w */
      for (y = 0; y < MIN(pbuffer->height, g->height); y++)
	{
	  guchar *ip;
	  guchar *buf;

	  ip = ppixels;
	  buf = pixel;
	  for (x = 0; x < MIN(pbuffer->width, g->width); x++)
	    {
	      buf[0] = buf[0] + (((ip[0] - buf[0]) * ip[3] + 0x80) >> 8);
	      buf[1] = buf[1] + (((ip[1] - buf[1]) * ip[3] + 0x80) >> 8);
	      buf[2] = buf[2] + (((ip[2] - buf[2]) * ip[3] + 0x80) >> 8);
	      buf[3] = buf[3] + (((ip[3] - buf[3]) * ip[3] + 0x80) >> 8);
	      ip += 4;
	      buf += 4;
	    }
	  ppixels += pbuffer->rowstride;
	  pixel += g->rowstride;
	}
    }

  ie_selection_destroy (app);
  ie_app_dirty (app);
  ie_update_grid (app);
}

void
edit_paste_into_cb (GtkWidget *widget,
		    gpointer data)
{
  GtkWidget *app = GTK_WIDGET (data);
  IEPasteBuffer *pbuffer;
  IESelection *selection;
  Grid *g;
  guchar *pixels, *pixel;
  guchar *ppixels;
  gboolean keep_trans = FALSE;
  gpointer kt;
  gint y, x;

  if (gtk_paste_buffer_get_size (GTK_PASTE_BUFFER (paste_buffer)) == 0)
    {
      gnome_error_dialog_parented (_("Paste Buffer is empty"), GTK_WINDOW (app));
      return;
    }

  /* We don't want to create a selection, just use one if it already exists */
  selection = gtk_object_get_data (GTK_OBJECT (app), "selection");
  if (selection == NULL)
    {
      gnome_error_dialog_parented (_("No area is selected."),
				   GTK_WINDOW (app));
      return;
    }

  kt = gtk_object_get_data (paste_buffer, "keep-trans");
  if (kt)
    keep_trans = GPOINTER_TO_INT (kt);

  pbuffer = gtk_paste_buffer_get (GTK_PASTE_BUFFER (paste_buffer), 
				  GTK_PASTE_BUFFER (paste_buffer)->selection);

  g = gtk_object_get_data (GTK_OBJECT (app), "grid");

  pixels = gdk_pixbuf_get_pixels (g->pixbuf);
  pixel = pixels + (selection->y * g->width * 4) + (selection->x * 4);

  ppixels = gdk_pixbuf_get_pixels (pbuffer->pixbuf);

  if (!keep_trans)
    {
      /* Fast */
      for (y = 0; y < MIN(pbuffer->height, selection->height); y++)
	{
	  memcpy (pixel, ppixels, MIN(pbuffer->width, selection->width) * 4);
	  
	  ppixels += pbuffer->rowstride;
	  pixel += g->rowstride;
	}  
    }
  else
    {
      /* S...l....oooooooooo.........w */
      for (y = 0; y < MIN(pbuffer->height, selection->height); y++)
	{
	  guchar *ip;
	  guchar *buf;

	  ip = ppixels;
	  buf = pixel;
	  for (x = 0; x < MIN(pbuffer->width, selection->width); x++)
	    {
	      buf[0] = buf[0] + (((ip[0] - buf[0]) * ip[3] + 0x80) >> 8);
	      buf[1] = buf[1] + (((ip[1] - buf[1]) * ip[3] + 0x80) >> 8);
	      buf[2] = buf[2] + (((ip[2] - buf[2]) * ip[3] + 0x80) >> 8);
	      buf[3] = buf[3] + (((ip[3] - buf[3]) * ip[3] + 0x80) >> 8);
	      ip += 4;
	      buf += 4;
	    }
	  ppixels += pbuffer->rowstride;
	  pixel += g->rowstride;
	}
    }

  ie_selection_destroy (app);
  ie_app_dirty (app);
  ie_update_grid (app);
}

void
edit_paste_at (Grid *g,
	       gint p_x,
	       gint p_y)
{
  GtkWidget *app = g->app;
  IEPasteBuffer *pbuffer;
  guchar *pixels, *pixel;
  guchar *ppixels;
  gboolean keep_trans = FALSE;
  gpointer kt;
  gint y, x;

  if (gtk_paste_buffer_get_size (GTK_PASTE_BUFFER (paste_buffer)) == 0)
    {
      gnome_error_dialog_parented (_("Paste Buffer is empty"), GTK_WINDOW (app));
      return;
    }

  kt = gtk_object_get_data (paste_buffer, "keep-trans");
  if (kt)
    keep_trans = GPOINTER_TO_INT (kt);

  pbuffer = gtk_paste_buffer_get (GTK_PASTE_BUFFER (paste_buffer), 
				  GTK_PASTE_BUFFER (paste_buffer)->selection);

  g = gtk_object_get_data (GTK_OBJECT (app), "grid");

  pixels = gdk_pixbuf_get_pixels (g->pixbuf);
  pixel = pixels + (p_y * g->width * 4) + (p_x * 4);

  ppixels = gdk_pixbuf_get_pixels (pbuffer->pixbuf);

  if (!keep_trans)
    {
      /* Fast */
      for (y = 0; y < MIN(pbuffer->height, g->height - p_y); y++)
	{
	  memcpy (pixel, ppixels, MIN(pbuffer->width, g->width - p_x) * 4);
	  
	  ppixels += pbuffer->rowstride;
	  pixel += g->rowstride;
	}  
    }
  else
    {
      /* S...l....oooooooooo.........w */
      for (y = 0; y < MIN(pbuffer->height, g->height - p_y); y++)
	{
	  guchar *ip;
	  guchar *buf;

	  ip = ppixels;
	  buf = pixel;
	  for (x = 0; x < MIN(pbuffer->width, g->width - p_x); x++)
	    {
	      buf[0] = buf[0] + (((ip[0] - buf[0]) * ip[3] + 0x80) >> 8);
	      buf[1] = buf[1] + (((ip[1] - buf[1]) * ip[3] + 0x80) >> 8);
	      buf[2] = buf[2] + (((ip[2] - buf[2]) * ip[3] + 0x80) >> 8);
	      buf[3] = buf[3] + (((ip[3] - buf[3]) * ip[3] + 0x80) >> 8);
	      ip += 4;
	      buf += 4;
	    }
	  ppixels += pbuffer->rowstride;
	  pixel += g->rowstride;
	}
    }

  ie_selection_destroy (app);
  ie_app_dirty (app);
  ie_update_grid (app);
}
